from typing import *


class Response:
    """A base class to be inherited by all other response types."""

    def to_dict(self) -> dict:
        """Prepare the Response to be sent by converting it to a JSONable :py:class:`dict`."""
        return {
            "type": self.__class__.__name__,
            **self.__dict__
        }

    def __eq__(self, other):
        if isinstance(other, Response):
            return self.to_dict() == other.to_dict()
        return False

    @classmethod
    def from_dict(cls, d: dict) -> "Response":
        """Recreate the response from a received :py:class:`dict`."""
        # Ignore type in dict
        del d["type"]
        # noinspection PyArgumentList
        return cls(**d)


class ResponseSuccess(Response):
    """A response to a successful :py:class:`Request`."""

    def __init__(self, data: Optional[dict] = None):
        if data is None:
            self.data = {}
        else:
            self.data = data

    def __repr__(self):
        return f"{self.__class__.__qualname__}(data={self.data})"


class ResponseFailure(Response):
    """A response to a invalid :py:class:`Request`."""

    def __init__(self, name: str, description: str, extra_info: Optional[dict] = None):
        self.name: str = name
        self.description: str = description
        self.extra_info: Optional[dict] = extra_info

    def __repr__(self):
        return f"{self.__class__.__qualname__}(name={self.name}, description={self.description}, extra_info={self.extra_info})"



"""The subpackage providing all functions and classes related to music playback on Discord.

It requires the both the ``bard`` and ``discord`` extras to be installed.

You can install them with: ::

    pip install royalnet[bard,discord]

"""

from .ytdldiscord import YtdlDiscord
from .fileaudiosource import FileAudioSource

__all__ = [
    "YtdlDiscord",
    "FileAudioSource",
]



import datetime
import secrets
from sqlalchemy import *
from sqlalchemy.orm import *
from sqlalchemy.ext.declarative import declared_attr
import royalnet.utils as ru


class Token:
    __tablename__ = "tokens"

    @declared_attr
    def token(self):
        return Column(String, primary_key=True)

    @declared_attr
    def user_id(self):
        return Column(Integer, ForeignKey("users.uid"), nullable=False)

    @declared_attr
    def user(self):
        return relationship("User", backref="tokens")

    @declared_attr
    def expiration(self):
        return Column(DateTime, nullable=False)

    @property
    def expired(self):
        return datetime.datetime.now() > self.expiration

    @expired.setter
    def expired(self, value):
        if value is True:
            self.expiration = datetime.datetime.fromtimestamp(0)
        else:
            raise ValueError("'expired' can only be set to True.")

    @classmethod
    def generate(cls, alchemy, user, expiration_delta: datetime.timedelta):
        # noinspection PyArgumentList
        TokenT = alchemy.get(cls)
        token = TokenT(user=user, expiration=datetime.datetime.now() + expiration_delta, token=secrets.token_urlsafe())
        return token

    def json(self) -> dict:
        return {
            "user": self.user.json(),
            "token": self.token,
            "expiration": self.expiration.isoformat()
        }

    @classmethod
    async def find(cls, alchemy, session, token: str) -> "Token":
        return await ru.asyncify(session.query(alchemy.get(cls)).filter_by(token=token).one_or_none)



from starlette.responses import *
from royalnet.utils import *
from royalnet.backpack.tables import *
from royalnet.constellation.api import *


class ApiAliasListStar(ApiStar):
    path = "/api/alias/list/v1"

    summary = "Get all aliases of the specified user."

    tags = ["alias"]

    parameters = {
        "user_id": "The id of the user to get the aliases of."
    }

    async def api(self, data: ApiData) -> JSON:
        aliases: typing.List[Alias] = await asyncify(
            data.session
                .query(self.alchemy.get(Alias))
                .filter_by(user_id=data["user_id"])
                .all
        )
        return [alias.alias for alias in aliases]



from typing import *
import asyncio as aio
import uuid
import functools
import logging
import websockets
from .package import Package
from .request import Request
from .response import Response, ResponseSuccess, ResponseFailure
from .broadcast import Broadcast
from .errors import ConnectionClosedError, InvalidServerResponseError
from .config import Config


log = logging.getLogger(__name__)


class PendingRequest:
    def __init__(self, *, loop: aio.AbstractEventLoop = None):
        if loop is None:
            self.loop = aio.get_event_loop()
        else:
            self.loop = loop
        self.event: aio.Event = aio.Event(loop=loop)
        self.data: Optional[dict] = None

    def __repr__(self):
        if self.event.is_set():
            return f"<{self.__class__.__qualname__}: {self.data.__class__.__name__}>"
        return f"<{self.__class__.__qualname__}>"

    def set(self, data):
        self.data = data
        self.event.set()


def requires_connection(func):
    @functools.wraps(func)
    async def new_func(self, *args, **kwargs):
        await self.connect_event.wait()
        return await func(self, *args, **kwargs)
    return new_func


def requires_identification(func):
    @functools.wraps(func)
    async def new_func(self, *args, **kwargs):
        await self.identify_event.wait()
        return await func(self, *args, **kwargs)
    return new_func


class Link:
    def __init__(self, config: Config, request_handler, *,
                 loop: aio.AbstractEventLoop = None):
        self.config: Config = config
        self.nid: str = str(uuid.uuid4())
        self.websocket: Optional["websockets.WebSocketClientProtocol"] = None
        self.request_handler: Callable[[Union[Request, Broadcast]],
                                       Awaitable[Response]] = request_handler
        self._pending_requests: Dict[str, PendingRequest] = {}
        if loop is None:
            self._loop = aio.get_event_loop()
        else:
            self._loop = loop
        self.error_event: aio.Event = aio.Event(loop=self._loop)
        self.connect_event: aio.Event = aio.Event(loop=self._loop)
        self.identify_event: aio.Event = aio.Event(loop=self._loop)

    def __repr__(self):
        if self.identify_event.is_set():
            return f"<{self.__class__.__qualname__} (identified)>"
        elif self.connect_event.is_set():
            return f"<{self.__class__.__qualname__} (connected)>"
        elif self.error_event.is_set():
            return f"<{self.__class__.__qualname__} (error)>"
        else:
            return f"<{self.__class__.__qualname__} (disconnected)>"

    async def connect(self):
        """Connect to the :class:`Server` at :attr:`.config.url`."""
        log.debug(f"Connecting to Herald Server at {self.config.url}...")
        self.websocket = await websockets.connect(self.config.url, loop=self._loop)
        self.connect_event.set()
        log.debug(f"Connected!")

    @requires_connection
    async def receive(self) -> Package:
        """Recieve a :py:class:`Package` from the :py:class:`Server`.

        Raises:
            :exc:`ConnectionClosedError` if the connection is closed."""
        try:
            jbytes: bytes = await self.websocket.recv()
            package: Package = Package.from_json_bytes(jbytes)
        except websockets.ConnectionClosed:
            self.error_event.set()
            self.connect_event.clear()
            self.identify_event.clear()
            log.warning(f"Herald Server connection closed: {self.config.url}")
            # What to do now? Let's just reraise.
            raise ConnectionClosedError()
        if self.identify_event.is_set() and package.destination != self.nid:
            raise InvalidServerResponseError("Package is not addressed to this NetworkLink.")
        log.debug(f"Received package: {package}")
        return package

    @requires_connection
    async def identify(self) -> None:
        log.debug(f"Identifying...")
        await self.websocket.send(f"Identify {self.nid}:{self.config.name}:{self.config.secret}")
        response: Package = await self.receive()
        if not response.source == "<server>":
            raise InvalidServerResponseError("Received a non-service package before identification.")
        if "type" not in response.data:
            raise InvalidServerResponseError("Missing 'type' in response data")
        if response.data["type"] == "error":
            raise ConnectionClosedError(f"Identification error: {response.data['type']}")
        assert response.data["type"] == "success"
        self.identify_event.set()
        log.debug(f"Identified successfully!")

    @requires_identification
    async def send(self, package: Package):
        """Send a package to the :class:`Server`."""
        log.debug(f"Trying to send package: {package}")
        try:
            jbytes = package.to_json_bytes()
        except TypeError as e:
            log.fatal(f"Could not send package: {' '.join(e.args)}")
            raise
        await self.websocket.send(jbytes)
        log.debug(f"Sent package: {package}")

    @requires_identification
    async def broadcast(self, destination: str, broadcast: Broadcast) -> None:
        package = Package(broadcast.to_dict(), source=self.nid, destination=destination)
        await self.send(package)
        log.debug(f"Sent broadcast to {destination}: {broadcast}")

    @requires_identification
    async def request(self, destination: str, request: Request) -> Response:
        if destination.startswith("*"):
            raise ValueError("requests cannot have multiple destinations")
        package = Package(request.to_dict(), source=self.nid, destination=destination)
        request = PendingRequest(loop=self._loop)
        self._pending_requests[package.source_conv_id] = request
        await self.send(package)
        log.debug(f"Sent request to {destination}: {request}")
        await request.event.wait()
        if request.data["type"] == "ResponseSuccess":
            response: Response = ResponseSuccess.from_dict(request.data)
        elif request.data["type"] == "ResponseFailure":
            response: Response = ResponseFailure.from_dict(request.data)
        else:
            raise TypeError("Unknown response type")
        log.debug(f"Received from {destination}: {request} -> {response}")
        return response

    async def run(self):
        """Blockingly run the Link."""
        log.debug(f"Running link: {self.config.name}")
        if self.error_event.is_set():
            raise ConnectionClosedError("RoyalnetLinks can't be rerun after an error.")
        while True:
            if not self.connect_event.is_set():
                await self.connect()
            if not self.identify_event.is_set():
                await self.identify()
            package: Package = await self.receive()
            # Package is a response
            if package.destination_conv_id in self._pending_requests:
                request = self._pending_requests[package.destination_conv_id]
                request.set(package.data)
                continue
            # Package is a request
            elif package.data["msg_type"] == "Request":
                log.debug(f"Received request {package.source_conv_id}: {package}")
                response: Response = await self.request_handler(Request.from_dict(package.data))
                response_package: Package = package.reply(response.to_dict())
                await self.send(response_package)
                log.debug(f"Replied to request {response_package.source_conv_id}: {response_package}")
            # Package is a broadcast
            elif package.data["msg_type"] == "Broadcast":
                log.debug(f"Received broadcast {package.source_conv_id}: {package}")
                await self.request_handler(Broadcast.from_dict(package.data))



import royalnet.version as rv
from royalnet.constellation.api import *
import royalnet.utils as ru


class ApiRoyalnetVersionStar(ApiStar):
    path = "/api/royalnet/version/v1"

    summary = "Get the current Royalnet version."

    tags = ["royalnet"]

    async def api(self, data: ApiData) -> ru.JSON:
        return {
            "semantic": rv.semantic
        }



class ApiError(Exception):
    pass


class NotFoundError(ApiError):
    pass


class ForbiddenError(ApiError):
    pass


class BadRequestError(ApiError):
    pass


class ParameterError(BadRequestError):
    pass


class MissingParameterError(ParameterError):
    pass


class InvalidParameterError(ParameterError):
    pass


class NotImplementedError(ApiError):
    pass


class UnsupportedError(NotImplementedError):
    pass



import asyncio
import threading
import logging
from typing import Optional
from .errors import *
from .playable import Playable
try:
    import discord
except ImportError:
    discord = None

log = logging.getLogger(__name__)


class VoicePlayer:
    def __init__(self, *, loop: Optional[asyncio.AbstractEventLoop] = None):
        self.voice_client: Optional["discord.VoiceClient"] = None
        self.playing: Optional[Playable] = None
        if loop is None:
            self.loop: asyncio.AbstractEventLoop = asyncio.get_event_loop()
        else:
            self.loop = loop
        self._playback_ended_event: threading.Event = threading.Event()

    async def connect(self, channel: "discord.VoiceChannel") -> "discord.VoiceClient":
        """Connect the :class:`VoicePlayer` to a :class:`discord.VoiceChannel`, creating a :class:`discord.VoiceClient`
        that handles the connection.

        Args:
            channel: The :class:`discord.VoiceChannel` to connect into.

        Returns:
            The created :class:`discord.VoiceClient`.
            (It will be stored in :attr:`VoicePlayer.voice_client` anyways!)

        Raises:
            PlayerAlreadyConnectedError:
            DiscordTimeoutError:
            GuildAlreadyConnectedError:
            OpusNotLoadedError:
        """
        if self.voice_client is not None and self.voice_client.is_connected():
            raise PlayerAlreadyConnectedError()
        log.debug(f"Connecting to: {channel}")
        try:
            self.voice_client = await channel.connect(reconnect=False, timeout=3)
        except asyncio.TimeoutError:
            raise DiscordTimeoutError()
        except discord.ClientException:
            raise GuildAlreadyConnectedError()
        except discord.opus.OpusNotLoaded:
            raise OpusNotLoadedError()
        return self.voice_client

    async def disconnect(self) -> None:
        """Disconnect the :class:`VoicePlayer` from the channel where it is currently connected, and set
        :attr:`.voice_client` to :const:`None`.

        Raises:
            PlayerNotConnectedError:
        """
        if self.voice_client is None or not self.voice_client.is_connected():
            raise PlayerNotConnectedError()
        log.debug(f"Disconnecting...")
        await self.voice_client.disconnect(force=True)
        self.voice_client = None

    async def move(self, channel: "discord.VoiceChannel"):
        """Move the :class:`VoicePlayer` to a different channel.

        This requires the :class:`VoicePlayer` to already be connected, and for the passed :class:`discord.VoiceChannel`
        to be in the same :class:`discord.Guild` of the :class:`VoicePlayer`."""
        if self.voice_client is None or not self.voice_client.is_connected():
            raise PlayerNotConnectedError()
        if self.voice_client.guild != channel.guild:
            raise ValueError("Can't move between two guilds.")
        log.debug(f"Moving to: {channel}")
        await self.voice_client.move_to(channel)

    async def start(self):
        """Start playing music on the :class:`discord.VoiceClient`.

        Info:
            Doesn't pass any ``*args`` or ``**kwargs`` to the :class:`Playable`.
        """
        if self.voice_client is None or not self.voice_client.is_connected():
            raise PlayerNotConnectedError()
        if self.voice_client.is_playing():
            raise PlayerAlreadyPlayingError()
        log.debug("Getting next AudioSource...")
        next_source: Optional["discord.AudioSource"] = await self.playing.next()
        if next_source is None:
            log.debug(f"Next source would be None, stopping here...")
            return
        log.debug(f"Next: {next_source}")
        self.voice_client.play(next_source, after=self._playback_ended)
        self.loop.create_task(self._playback_check())

    async def _playback_check(self):
        while True:
            if self._playback_ended_event.is_set():
                self._playback_ended_event.clear()
                await self.start()
                break
            await asyncio.sleep(1)

    def _playback_ended(self, error=None):
        if error is not None:
            # TODO: catch with Sentry
            log.error(error)
            return
        self._playback_ended_event.set()

    async def change_playing(self, value: Playable):
        await self.playing.destroy()
        self.playing = value



import click
import multiprocessing
import toml
import logging
import royalnet.constellation as rc
import royalnet.serf.telegram as rst
import royalnet.serf.discord as rsd
import royalnet.serf.matrix as rsm
import royalnet.utils as ru
import royalnet.herald as rh

try:
    import coloredlogs
except ImportError:
    coloredlogs = None


log = logging.getLogger(__name__)


@click.command()
@click.option("-c", "--config-filename", default="./config.toml", type=click.Path(exists=True),
              help="The filename of the Royalnet configuration file.")
def run(config_filename: str):
    # Read the configuration file
    with open(config_filename, "r") as t:
        config: dict = toml.load(t)

    ru.init_logging(config["Logging"])

    if config["Sentry"] is None or not config["Sentry"]["enabled"]:
        log.info("Sentry: disabled")
    else:
        try:
            ru.init_sentry(config["Sentry"])
        except ImportError:
            log.info("Sentry: not installed")

    # Herald Server
    herald_cfg = None
    herald_process = None
    if config["Herald"]["Local"]["enabled"]:
        # Create a Herald server
        herald_server = rh.Server(rh.Config.from_config(name="<server>", **config["Herald"]["Local"]))
        # Run the Herald server on a new process
        herald_process = multiprocessing.Process(name="Herald.Local",
                                                 target=herald_server.run_blocking,
                                                 daemon=True,
                                                 kwargs={
                                                     "logging_cfg": config["Logging"]
                                                 })
        herald_process.start()
        herald_cfg = config["Herald"]["Local"]
        log.info("Herald: Enabled (Local)")
    elif config["Herald"]["Remote"]["enabled"]:
        log.info("Herald: Enabled (Remote)")
        herald_cfg = config["Herald"]["Remote"]
    else:
        log.info("Herald: Disabled")

    # Serfs
    telegram_process = None
    if "Telegram" in config["Serfs"] and config["Serfs"]["Telegram"]["enabled"]:
        telegram_process = multiprocessing.Process(name="Serf.Telegram",
                                                   target=rst.TelegramSerf.run_process,
                                                   daemon=True,
                                                   kwargs={
                                                       "alchemy_cfg": config["Alchemy"],
                                                       "herald_cfg": herald_cfg,
                                                       "packs_cfg": config["Packs"],
                                                       "sentry_cfg": config["Sentry"],
                                                       "logging_cfg": config["Logging"],
                                                       "serf_cfg": config["Serfs"]["Telegram"],
                                                   })
        telegram_process.start()
        log.info("Serf.Telegram: Started")
    else:
        log.info("Serf.Telegram: Disabled")

    discord_process = None
    if "Discord" in config["Serfs"] and config["Serfs"]["Discord"]["enabled"]:
        discord_process = multiprocessing.Process(name="Serf.Discord",
                                                  target=rsd.DiscordSerf.run_process,
                                                  daemon=True,
                                                  kwargs={
                                                      "alchemy_cfg": config["Alchemy"],
                                                      "herald_cfg": herald_cfg,
                                                      "packs_cfg": config["Packs"],
                                                      "sentry_cfg": config["Sentry"],
                                                      "logging_cfg": config["Logging"],
                                                      "serf_cfg": config["Serfs"]["Discord"],
                                                  })
        discord_process.start()
        log.info("Serf.Discord: Started")
    else:
        log.info("Serf.Discord: Disabled")

    matrix_process = None
    if "Matrix" in config["Serfs"] and config["Serfs"]["Matrix"]["enabled"]:
        matrix_process = multiprocessing.Process(name="Serf.Matrix",
                                                 target=rsm.MatrixSerf.run_process,
                                                 daemon=True,
                                                 kwargs={
                                                     "alchemy_cfg": config["Alchemy"],
                                                     "herald_cfg": herald_cfg,
                                                     "packs_cfg": config["Packs"],
                                                     "sentry_cfg": config["Sentry"],
                                                     "logging_cfg": config["Logging"],
                                                     "serf_cfg": config["Serfs"]["Matrix"],
                                                 })
        matrix_process.start()
        log.info("Serf.Matrix: Started")
    else:
        log.info("Serf.Matrix: Disabled")

    # Constellation
    constellation_process = None
    if config["Constellation"]["enabled"]:
        constellation_process = multiprocessing.Process(name="Constellation",
                                                        target=rc.Constellation.run_process,
                                                        daemon=True,
                                                        kwargs={
                                                            "alchemy_cfg": config["Alchemy"],
                                                            "herald_cfg": herald_cfg,
                                                            "packs_cfg": config["Packs"],
                                                            "sentry_cfg": config["Sentry"],
                                                            "logging_cfg": config["Logging"],
                                                            "constellation_cfg": config["Constellation"],
                                                        })
        constellation_process.start()
        log.info("Constellation: Started")
    else:
        log.info("Constellation: Disabled")

    log.info("All processes started!")
    if constellation_process is not None:
        constellation_process.join()
    if telegram_process is not None:
        telegram_process.join()
    if discord_process is not None:
        discord_process.join()
    if matrix_process is not None:
        matrix_process.join()
    if herald_process is not None:
        herald_process.join()


if __name__ == "__main__":
    run()



"""The subpackage providing all classes related to music files.

It requires the ``bard`` extra to be installed (the :mod:`ffmpeg_python`, :mod:`youtube_dl` and :mod:`eyed3` packages).

You can install it with: ::

    pip install royalnet[bard]

"""

try:
    import ffmpeg
    import youtube_dl
    import eyed3
except ImportError:
    raise ImportError("The `bard` extra is not installed. Please install it with `pip install royalnet[bard]`.")

from .ytdlinfo import YtdlInfo
from .ytdlfile import YtdlFile
from .errors import BardError, YtdlError, NotFoundError, MultipleFilesError

__all__ = [
    "YtdlInfo",
    "YtdlFile",
    "BardError",
    "YtdlError",
    "NotFoundError",
    "MultipleFilesError",
]



from typing import *
import asyncio as aio
import logging
import importlib
import uvicorn
import starlette.applications
import royalnet.alchemy as ra
import royalnet.herald as rh
import royalnet.utils as ru
import royalnet.commands as rc
from .pagestar import PageStar
from ..utils import init_logging


log = logging.getLogger(__name__)

UVICORN_LOGGING_CONFIG = {
    "version": 1,
    "disable_existing_loggers": True,
    "formatters": {},
    "handlers": {},
    "loggers": {},
}


class Constellation:
    """The class that represents the webserver.

    It runs multiple :class:`Star`, which represent the routes of the website.

    It also handles the :class:`Alchemy` connection, and Herald connections too."""
    def __init__(self,
                 alchemy_cfg: Dict[str, Any],
                 herald_cfg: Dict[str, Any],
                 packs_cfg: Dict[str, Any],
                 constellation_cfg: Dict[str, Any],
                 logging_cfg: Dict[str, Any]
                 ):
        # Import packs
        pack_names = packs_cfg["active"]
        packs = {}
        for pack_name in pack_names:
            log.debug(f"Importing pack: {pack_name}")
            try:
                packs[pack_name] = {
                    "commands": importlib.import_module(f"{pack_name}.commands"),
                    "events": importlib.import_module(f"{pack_name}.events"),
                    "stars": importlib.import_module(f"{pack_name}.stars"),
                    "tables": importlib.import_module(f"{pack_name}.tables"),
                }
            except ImportError as e:
                log.error(f"Error during the import of {pack_name}: {e}")
        log.info(f"Packs: {len(packs)} imported")

        self.alchemy = None
        """The :class:`~ra.Alchemy` of this Constellation."""

        # Alchemy
        if ra.Alchemy is None:
            log.info("Alchemy: not installed")
        elif not alchemy_cfg["enabled"]:
            log.info("Alchemy: disabled")
        else:
            # Find all tables
            tables = set()
            for pack in packs.values():
                try:
                    tables = tables.union(pack["tables"].available_tables)
                except AttributeError:
                    log.warning(f"Pack `{pack}` does not have the `available_tables` attribute.")
                    continue
            # Create the Alchemy
            self.alchemy = ra.Alchemy(alchemy_cfg["database_url"], tables)
            log.info(f"Alchemy: {self.alchemy}")

        # Logging
        self._logging_cfg: Dict[str, Any] = logging_cfg
        """The logging config for the :class:`Constellation` is stored to initialize the logger when the first page is
        requested, as disabling the :mod:`uvicorn` logging also disables all logging in the process in general."""

        # Herald
        self.herald: Optional[rh.Link] = None
        """The :class:`Link` object connecting the :class:`Constellation` to the rest of the herald network.
        As is the case with the logging module, it will be started on the first request received by the 
        :class:`Constellation`, as the event loop won't be available before that."""

        self._herald_cfg: Dict[str, Any] = herald_cfg
        """The herald config for the :class:`Constellation` is stored to initialize the :class:`rh.Herald` later."""

        self.herald_task: Optional[aio.Task] = None
        """A reference to the :class:`aio.Task` that runs the :class:`rh.Link`."""

        self.Interface: Type[rc.CommandInterface] = self.interface_factory()
        """The :class:`~rc.CommandInterface` class of this :class:`Constellation`."""

        self.events: Dict[str, rc.Event] = {}
        """A dictionary containing all :class:`~rc.Event` that can be handled by this :class:`Constellation`."""

        self.starlette = starlette.applications.Starlette(debug=__debug__)
        """The :class:`~starlette.Starlette` app."""

        self.stars: List[PageStar] = []
        """A list of all the :class:`PageStar` registered to this :class:`Constellation`."""

        # Register Events
        for pack_name in packs:
            pack = packs[pack_name]
            pack_cfg = packs_cfg.get(pack_name, {})
            try:
                events = pack["events"].available_events
            except AttributeError:
                log.warning(f"Pack `{pack}` does not have the `available_events` attribute.")
            else:
                self.register_events(events, pack_cfg)
        log.info(f"Events: {len(self.events)} events")

        if rh.Link is None:
            log.info("Herald: not installed")
        elif not herald_cfg["enabled"]:
            log.info("Herald: disabled")
        else:
            log.info(f"Herald: will be enabled on first request")

        # Register PageStars and ExceptionStars
        for pack_name in packs:
            pack = packs[pack_name]
            pack_cfg = packs_cfg.get(pack_name, {})
            try:
                page_stars = pack["stars"].available_page_stars
            except AttributeError:
                log.warning(f"Pack `{pack}` does not have the `available_page_stars` attribute.")
            else:
                self.register_page_stars(page_stars, pack_cfg)
        log.info(f"PageStars: {len(self.starlette.routes)} stars")
        log.info(f"ExceptionStars: {len(self.starlette.exception_handlers)} stars")

        self.running: bool = False
        """Is the :class:`Constellation` server currently running?"""

        self.address: str = constellation_cfg["address"]
        """The address that the :class:`Constellation` will bind to when run."""

        self.port: int = constellation_cfg["port"]
        """The port on which the :class:`Constellation` will listen for connection on."""

        self.loop: Optional[aio.AbstractEventLoop] = None
        """The event loop of the :class:`Constellation`. 
        
        Because of how :mod:`uvicorn` runs, it will stay :const:`None` until the first page is requested."""

    # TODO: is this a good idea?
    def interface_factory(self) -> Type[rc.CommandInterface]:
        """Create the :class:`rc.CommandInterface` class for the :class:`Constellation`."""

        # noinspection PyMethodParameters
        class GenericInterface(rc.CommandInterface):
            alchemy: ra.Alchemy = self.alchemy
            constellation = self

            async def call_herald_event(ci, destination: str, event_name: str, **kwargs) -> Dict:
                """Send a :class:`royalherald.Request` to a specific destination, and wait for a
                :class:`royalherald.Response`."""
                if self.herald is None:
                    raise rc.UnsupportedError("`royalherald` is not enabled on this serf.")
                request: rh.Request = rh.Request(handler=event_name, data=kwargs)
                response: rh.Response = await self.herald.request(destination=destination, request=request)
                if isinstance(response, rh.ResponseFailure):
                    if response.name == "no_event":
                        raise rc.ProgramError(f"There is no event named {event_name} in {destination}.")
                    elif response.name == "error_in_event":
                        if response.extra_info["type"] == "CommandError":
                            raise rc.CommandError(response.extra_info["message"])
                        elif response.extra_info["type"] == "UserError":
                            raise rc.UserError(response.extra_info["message"])
                        elif response.extra_info["type"] == "InvalidInputError":
                            raise rc.InvalidInputError(response.extra_info["message"])
                        elif response.extra_info["type"] == "UnsupportedError":
                            raise rc.UnsupportedError(response.extra_info["message"])
                        elif response.extra_info["type"] == "ConfigurationError":
                            raise rc.ConfigurationError(response.extra_info["message"])
                        elif response.extra_info["type"] == "ExternalError":
                            raise rc.ExternalError(response.extra_info["message"])
                        else:
                            raise rc.ProgramError(f"Invalid error in Herald event '{event_name}':\n"
                                                  f"[b]{response.extra_info['type']}[/b]\n"
                                                  f"{response.extra_info['message']}")
                    elif response.name == "unhandled_exception_in_event":
                        raise rc.ProgramError(f"Unhandled exception in Herald event '{event_name}':\n"
                                              f"[b]{response.extra_info['type']}[/b]\n"
                                              f"{response.extra_info['message']}")
                    else:
                        raise rc.ProgramError(f"Unknown response in Herald event '{event_name}':\n"
                                              f"[b]{response.name}[/b]"
                                              f"[p]{response}[/p]")
                elif isinstance(response, rh.ResponseSuccess):
                    return response.data
                else:
                    raise rc.ProgramError(f"Other Herald Link returned unknown response:\n"
                                          f"[p]{response}[/p]")

        return GenericInterface

    def init_herald(self, herald_cfg: Dict[str, Any]):
        """Create a :class:`rh.Link`."""
        herald_cfg["name"] = "constellation"
        self.herald: rh.Link = rh.Link(rh.Config.from_config(**herald_cfg), self.network_handler)

    async def network_handler(self, message: Union[rh.Request, rh.Broadcast]) -> rh.Response:
        try:
            event: rc.Event = self.events[message.handler]
        except KeyError:
            log.warning(f"No event for '{message.handler}'")
            return rh.ResponseFailure("no_event", f"This serf does not have any event for {message.handler}.")
        log.debug(f"Event called: {event.name}")
        if isinstance(message, rh.Request):
            try:
                response_data = await event.run(**message.data)
                return rh.ResponseSuccess(data=response_data)
            except Exception as e:
                ru.sentry_exc(e)
                return rh.ResponseFailure("exception_in_event",
                                          f"An exception was raised in the event for '{message.handler}'.",
                                          extra_info={
                                              "type": e.__class__.__qualname__,
                                              "message": str(e)
                                          })
        elif isinstance(message, rh.Broadcast):
            await event.run(**message.data)

    def register_events(self, events: List[Type[rc.Event]], pack_cfg: Dict[str, Any]):
        for SelectedEvent in events:
            # Create a new interface
            interface = self.Interface(config=pack_cfg)
            # Initialize the event
            try:
                event = SelectedEvent(interface)
            except Exception as e:
                log.error(f"Skipping: "
                          f"{SelectedEvent.__qualname__} - {e.__class__.__qualname__} in the initialization.")
                ru.sentry_exc(e)
                continue
            # Register the event
            if SelectedEvent.name in self.events:
                log.warning(f"Overriding (already defined): {SelectedEvent.__qualname__} -> {SelectedEvent.name}")
            else:
                log.debug(f"Registering: {SelectedEvent.__qualname__} -> {SelectedEvent.name}")
            self.events[SelectedEvent.name] = event

    def _first_page_check(self):
        if self.loop is None:
            self.loop = aio.get_running_loop()
            self.init_herald(self._herald_cfg)
            self.loop.create_task(self.herald.run())
            init_logging(self._logging_cfg)

    def _page_star_wrapper(self, page_star: PageStar):
        async def f(request):
            self._first_page_check()
            log.info(f"Running {page_star}")
            return await page_star.page(request)

        return page_star.path, f, page_star.methods

    def register_page_stars(self, page_stars: List[Type[PageStar]], pack_cfg: Dict[str, Any]):
        for SelectedPageStar in page_stars:
            log.debug(f"Registering: {SelectedPageStar.path} -> {SelectedPageStar.__qualname__}")
            try:
                page_star_instance = SelectedPageStar(interface=self.Interface(pack_cfg))
            except Exception as e:
                log.error(f"Skipping: "
                          f"{SelectedPageStar.__qualname__} - {e.__class__.__qualname__} in the initialization.")
                ru.sentry_exc(e)
                continue
            self.stars.append(page_star_instance)
            self.starlette.add_route(*self._page_star_wrapper(page_star_instance))

    def run_blocking(self):
        log.info(f"Running Constellation on https://{self.address}:{self.port}/...")
        loop: aio.AbstractEventLoop = aio.get_event_loop()
        self.running = True
        try:
            uvicorn.run(self.starlette, host=self.address, port=self.port, log_config=UVICORN_LOGGING_CONFIG)
        finally:
            self.running = False

    @classmethod
    def run_process(cls,
                    alchemy_cfg: Dict[str, Any],
                    herald_cfg: Dict[str, Any],
                    sentry_cfg: Dict[str, Any],
                    packs_cfg: Dict[str, Any],
                    constellation_cfg: Dict[str, Any],
                    logging_cfg: Dict[str, Any]):
        """Blockingly create and run the Constellation.

        This should be used as the target of a :class:`multiprocessing.Process`."""
        ru.init_logging(logging_cfg)

        if sentry_cfg is None or not sentry_cfg["enabled"]:
            log.info("Sentry: disabled")
        else:
            try:
                ru.init_sentry(sentry_cfg)
            except ImportError:
                log.info("Sentry: not installed")

        constellation = cls(alchemy_cfg=alchemy_cfg,
                            herald_cfg=herald_cfg,
                            packs_cfg=packs_cfg,
                            constellation_cfg=constellation_cfg,
                            logging_cfg=logging_cfg)

        # Run the server
        constellation.run_blocking()

    def __repr__(self):
        return f"<{self.__class__.__qualname__}: {'running' if self.running else 'inactive'}>"



import datetime
import royalnet.utils as ru
from royalnet.constellation.api import *
from ..tables.users import User
from ..tables.aliases import Alias
from ..tables.tokens import Token


class ApiLoginRoyalnetStar(ApiStar):
    path = "/api/login/royalnet/v1"

    methods = ["POST"]

    summary = "Login as a Royalnet user, creating a 7-day login token."

    parameters = {
        "username": "The name of the user you are logging in as.",
        "password": "The password of the user you are logging in as."
    }

    tags = ["login"]

    async def api(self, data: ApiData) -> ru.JSON:
        TokenT = self.alchemy.get(Token)
        UserT = self.alchemy.get(User)
        AliasT = self.alchemy.get(Alias)

        username = data["username"]
        password = data["password"]

        async with self.session_acm() as session:
            user: User = await ru.asyncify(session.query(UserT).filter_by(username=username).one_or_none)
            if user is None:
                raise NotFoundError("User not found")
            pswd_check = user.test_password(password)
            if not pswd_check:
                raise ApiError("Invalid password")
            token: Token = TokenT.generate(alchemy=self.alchemy, user=user, expiration_delta=datetime.timedelta(days=7))
            session.add(token)
            await ru.asyncify(session.commit)
            response = token.json()

        return response



from typing import *
import datetime
import royalnet.utils as ru
from royalnet.constellation.api import *
from ..tables.tokens import Token


class ApiTokenCreateStar(ApiStar):
    path = "/api/token/create/v1"

    methods = ["POST"]

    summary = "Create a new login token of any duration."

    parameters = {
        "duration": "The duration in seconds of the new token."
    }

    tags = ["token"]

    requires_auth = True

    async def api(self, data: ApiData) -> ru.JSON:
        user = await data.user()
        try:
            duration = int(data["duration"])
        except ValueError:
            raise InvalidParameterError("Duration is not a valid integer")
        new_token = Token.generate(self.alchemy, user, datetime.timedelta(seconds=duration))
        data.session.add(new_token)
        await data.session_commit()
        return new_token.json()



import contextlib
import logging
import asyncio as aio
import uuid
from typing import *
import royalnet.commands as rc
import royalnet.utils as ru
import royalnet.backpack as rb
from .escape import escape
from ..serf import Serf

try:
    import telegram
    import urllib3
    from telegram.utils.request import Request as TRequest
except ImportError:
    telegram = None
    urllib3 = None
    TRequest = None

try:
    from sqlalchemy.orm.session import Session
except ImportError:
    Session = None

log = logging.getLogger(__name__)


class TelegramSerf(Serf):
    """A Serf that connects to `Telegram <https://telegram.org/>`_ as a bot."""
    interface_name = "telegram"

    _identity_table = rb.tables.Telegram
    _identity_column = "tg_id"

    def __init__(self,
                 loop: aio.AbstractEventLoop,
                 alchemy_cfg: Dict[str, Any],
                 herald_cfg: Dict[str, Any],
                 sentry_cfg: Dict[str, Any],
                 packs_cfg: Dict[str, Any],
                 serf_cfg: Dict[str, Any],
                 **_):
        if telegram is None:
            raise ImportError("'telegram' extra is not installed")

        super().__init__(loop=loop,
                         alchemy_cfg=alchemy_cfg,
                         herald_cfg=herald_cfg,
                         sentry_cfg=sentry_cfg,
                         packs_cfg=packs_cfg,
                         serf_cfg=serf_cfg)

        self.client = telegram.Bot(serf_cfg["token"],
                                   request=TRequest(serf_cfg["pool_size"],
                                                    read_timeout=serf_cfg["read_timeout"]))
        """The :class:`telegram.Bot` instance that will be used from the Serf."""

        self.update_offset: int = -100
        """The current `update offset <https://core.telegram.org/bots/api#getupdates>`_."""

        self.key_callbacks: Dict[str, rc.KeyboardKey] = {}

        self.MessageData: Type[rc.CommandData] = self.message_data_factory()
        self.CallbackData: Type[rc.CommandData] = self.callback_data_factory()

    @staticmethod
    async def api_call(f: Callable, *args, **kwargs) -> Optional:
        """Call a :class:`telegram.Bot` method safely, without getting a mess of errors raised.

        The method may return None if it was decided that the call should be skipped."""
        while True:
            try:
                return await ru.asyncify(f, *args, **kwargs)
            except telegram.error.TimedOut as error:
                log.debug(f"Timed out during {f.__qualname__} (retrying immediatly): {error}")
                continue
            except telegram.error.NetworkError as error:
                log.debug(f"Network error during {f.__qualname__} (skipping): {error}")
                break
            except telegram.error.Unauthorized as error:
                log.info(f"Unauthorized to run {f.__qualname__} (skipping): {error}")
                break
            except telegram.error.RetryAfter as error:
                log.warning(f"Rate limited during {f.__qualname__} (retrying in 15s): {error}")
                await aio.sleep(15)
                continue
            except urllib3.exceptions.HTTPError as error:
                log.warning(f"urllib3 HTTPError during {f.__qualname__} (retrying in 15s): {error}")
                await aio.sleep(15)
                continue
            except Exception as error:
                log.error(f"{error.__class__.__qualname__} during {f} (skipping): {error}")
                ru.sentry_exc(error)
                break
        return None

    def interface_factory(self) -> Type[rc.CommandInterface]:
        # noinspection PyPep8Naming
        GenericInterface = super().interface_factory()

        # noinspection PyMethodParameters
        class TelegramInterface(GenericInterface):
            name = self.interface_name
            prefix = "/"

        return TelegramInterface

    def message_data_factory(self) -> Type[rc.CommandData]:
        # noinspection PyMethodParameters
        class TelegramMessageData(rc.CommandData):
            def __init__(data,
                         interface: rc.CommandInterface,
                         loop: aio.AbstractEventLoop,
                         message: telegram.Message):
                super().__init__(interface=interface, loop=loop)
                data.message: telegram.Message = message

            async def reply(data, text: str):
                await self.api_call(data.message.chat.send_message,
                                    escape(text),
                                    parse_mode="HTML",
                                    disable_web_page_preview=True)

            async def get_author(data, error_if_none=False):
                user: Optional[telegram.User] = data.message.from_user
                if user is None:
                    if error_if_none:
                        raise rc.CommandError("No command caller for this message")
                    return None
                query = data.session.query(self.master_table)
                for link in self.identity_chain:
                    query = query.join(link.mapper.class_)
                query = query.filter(self.identity_column == user.id)
                result = await ru.asyncify(query.one_or_none)
                if result is None and error_if_none:
                    raise rc.CommandError("Command caller is not registered")
                return result

            async def delete_invoking(data, error_if_unavailable=False) -> None:
                await self.api_call(data.message.delete)

            @contextlib.asynccontextmanager
            async def keyboard(data, text: str, keys: List[rc.KeyboardKey]):
                tg_rows = []
                key_uids = []
                for key in keys:
                    uid: str = str(uuid.uuid4())
                    key_uids.append(uid)
                    self.register_keyboard_key(uid, key)
                    tg_button: telegram.InlineKeyboardButton = telegram.InlineKeyboardButton(key.text,
                                                                                             callback_data=uid)
                    tg_row: List[telegram.InlineKeyboardButton] = [tg_button]
                    tg_rows.append(tg_row)
                tg_markup: telegram.InlineKeyboardMarkup = telegram.InlineKeyboardMarkup(tg_rows)
                message: telegram.Message = await self.api_call(data.message.chat.send_message,
                                                                escape(text),
                                                                parse_mode="HTML",
                                                                disable_web_page_preview=True,
                                                                reply_markup=tg_markup)
                yield message
                await self.api_call(message.edit_reply_markup, reply_markup=None)
                for uid in key_uids:
                    self.unregister_keyboard_key(uid)

        return TelegramMessageData

    def callback_data_factory(self) -> Type[rc.CommandData]:
        # noinspection PyMethodParameters
        class TelegramKeyboardData(rc.CommandData):
            def __init__(data,
                         interface: rc.CommandInterface,
                         loop: aio.AbstractEventLoop,
                         cbq: telegram.CallbackQuery):
                super().__init__(interface=interface, loop=loop)
                data.cbq: telegram.CallbackQuery = cbq

            async def reply(data, text: str):
                await self.api_call(data.cbq.answer,
                                    escape(text))

            async def get_author(data, error_if_none=False):
                user = data.cbq.from_user
                if user is None:
                    if error_if_none:
                        raise rc.CommandError("No command caller for this message")
                    return None
                query = data.session.query(self.master_table)
                for link in self.identity_chain:
                    query = query.join(link.mapper.class_)
                query = query.filter(self.identity_column == user.id)
                result = await ru.asyncify(query.one_or_none)
                if result is None and error_if_none:
                    raise rc.CommandError("Command caller is not registered")
                return result

        return TelegramKeyboardData

    async def answer_cbq(self, cbq, text, alert=False):
        await self.api_call(cbq.answer, text=text, show_alert=alert)

    async def handle_update(self, update: telegram.Update):
        """Delegate :class:`telegram.Update` handling to the correct message type submethod."""
        if update.message is not None:
            await self.handle_message(update.message)
        elif update.edited_message is not None:
            pass
        elif update.channel_post is not None:
            pass
        elif update.edited_channel_post is not None:
            pass
        elif update.inline_query is not None:
            pass
        elif update.chosen_inline_result is not None:
            pass
        elif update.callback_query is not None:
            await self.handle_callback_query(update.callback_query)
        elif update.shipping_query is not None:
            pass
        elif update.pre_checkout_query is not None:
            pass
        elif update.poll is not None:
            pass
        else:
            log.warning(f"Unknown update type: {update}")

    async def handle_message(self, message: telegram.Message):
        """What should be done when a :class:`telegram.Message` is received?"""
        text: str = message.text
        # Try getting the caption instead
        if text is None:
            text: str = message.caption
        # No text or caption, ignore the message
        if text is None:
            return
        # Skip non-command updates
        if not text.startswith("/"):
            return
        # Find and clean parameters
        command_text, *parameters = text.split(" ")
        command_name = command_text.replace(f"@{self.client.username}", "").lower()
        # Find the command
        try:
            command = self.commands[command_name]
        except KeyError:
            # Skip the message
            return
        # Send a typing notification
        await self.api_call(message.chat.send_action, telegram.ChatAction.TYPING)
        # Prepare data
        data = self.MessageData(interface=command.interface, loop=self.loop, message=message)
        # Call the command
        await self.call(command, data, parameters)

    async def handle_callback_query(self, cbq: telegram.CallbackQuery):
        uid = cbq.data
        if uid not in self.key_callbacks:
            await self.api_call(cbq.answer, text="⚠️ This keyboard has expired.", show_alert=True)
        key: rc.KeyboardKey = self.key_callbacks[uid]
        data: rc.CommandData = self.CallbackData(interface=key.interface, loop=self.loop, cbq=cbq)
        await self.press(key, data)

    def register_keyboard_key(self, identifier: str, key: rc.KeyboardKey):
        self.key_callbacks[identifier] = key

    def unregister_keyboard_key(self, identifier: str):
        del self.key_callbacks[identifier]

    async def run(self):
        await super().run()
        while True:
            # Get the latest 100 updates
            last_updates: List[telegram.Update] = await self.api_call(self.client.get_updates,
                                                                      offset=self.update_offset,
                                                                      timeout=60,
                                                                      read_latency=5.0)
            # Handle updates
            for update in last_updates:
                # TODO: don't lose the reference to the task
                # noinspection PyAsyncCall
                self.loop.create_task(self.handle_update(update))
            # Recalculate offset
            try:
                self.update_offset = last_updates[-1].update_id + 1
            except IndexError:
                pass



"""A :class:`Serf` implementation for Matrix.

It requires (obviously) the ``matrix`` extra to be installed.

Install it with: ::

    pip install royalnet[matrix]

"""

from .matrixserf import MatrixSerf
from .escape import escape

__all__ = [
    "MatrixSerf",
    "escape",
]



from typing import *
from json import JSONDecodeError
from abc import *
from starlette.requests import Request
from starlette.responses import JSONResponse
from ..pagestar import PageStar
from .jsonapi import api_error, api_success
from .apidata import ApiData
from .apierrors import *
import royalnet.utils as ru


class ApiStar(PageStar, ABC):
    summary: str = ""

    description: str = ""

    parameters: Dict[str, str] = {}

    tags: List[str] = []

    requires_auth: bool = False

    async def page(self, request: Request) -> JSONResponse:
        if request.query_params:
            data = request.query_params
        else:
            try:
                data = await request.json()
            except JSONDecodeError:
                data = {}
        apidata = ApiData(data, self)
        try:
            response = await self.api(apidata)
        except NotFoundError as e:
            return api_error(e, code=404)
        except ForbiddenError as e:
            return api_error(e, code=403)
        except NotImplementedError as e:
            return api_error(e, code=501)
        except BadRequestError as e:
            return api_error(e, code=400)
        except Exception as e:
            ru.sentry_exc(e)
            return api_error(e, code=500)
        else:
            return api_success(response)
        finally:
            await apidata.session_close()

    async def api(self, data: ApiData) -> ru.JSON:
        raise NotImplementedError()

    @classmethod
    def swagger(cls) -> ru.JSON:
        """Generate one or more swagger paths for this ApiStar."""
        result = {}
        for method in cls.methods:
            result[method.lower()] = {
                "operationId": cls.__name__,
                "summary": cls.summary,
                "description": cls.description,
                "responses": {
                    "200": {"description": "Success"},
                    "400": {"description": "Bad request"},
                    "403": {"description": "Forbidden"},
                    "404": {"description": "Not found"},
                    "500": {"description": "Serverside unhandled exception"},
                    "501": {"description": "Not yet implemented"}
                },
                "tags": cls.tags,
                "parameters": [{
                    "name": parameter,
                    "in": "query",
                    "description": cls.parameters[parameter],
                    "type": "string"
                } for parameter in cls.parameters]
            }
            if cls.requires_auth:
                result[method.lower()]["security"] = [{"RoyalnetLoginToken": ["logged_in"]}]
        return result



from typing import *
from contextlib import contextmanager, asynccontextmanager
from royalnet.utils import asyncify
from royalnet.alchemy.errors import TableNotFoundError
from sqlalchemy import create_engine
from sqlalchemy.engine import Engine
from sqlalchemy.schema import Table
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.ext.declarative.api import DeclarativeMeta, AbstractConcreteBase
from sqlalchemy.orm import sessionmaker


class Alchemy:
    """A wrapper around :mod:`sqlalchemy.orm` that allows the instantiation of multiple engines at once while maintaining
    a single declarative class for all of them."""

    def __init__(self, database_uri: str, tables: Set):
        """Create a new :class:`.Alchemy` object.

        Args:
            database_uri: The `database URI <https://docs.sqlalchemy.org/en/13/core/engines.html>`_  .
            tables: The :class:`set` of tables to be created and used in the selected database.
                    Check the tables submodule for more details.
        """
        if database_uri.startswith("sqlite"):
            raise NotImplementedError("sqlite databases aren't supported, as they can't be used in multithreaded"
                                      " applications")
        self._engine: Engine = create_engine(database_uri)
        self._Base = declarative_base(bind=self._engine)
        self.Session: sessionmaker = sessionmaker(bind=self._engine)
        self._tables: Dict[str, Table] = {}
        for table in tables:
            name = table.__name__
            assert self._tables.get(name) is None
            assert isinstance(name, str)
            # noinspection PyTypeChecker
            bound_table: Table = type(name, (self._Base, table), {})
            self._tables[name] = bound_table
        self._Base.metadata.create_all()

    def get(self, table: Union[str, type]) -> DeclarativeMeta:
        """Get the table with a specified name or class.

        Args:
            table: The table name or table class you want to get.

        Raises:
            TableNotFoundError: if the requested table was not found."""
        if isinstance(table, str):
            result = self._tables.get(table)
            if result is None:
                raise TableNotFoundError(f"Table '{table}' isn't present in this Alchemy instance")
            return result
        elif isinstance(table, type):
            name = table.__name__
            result = self._tables.get(name)
            if result is None:
                raise TableNotFoundError(f"Table '{table}' isn't present in this Alchemy instance")
            return result
        else:
            raise TypeError(f"Can't get tables with objects of type '{table.__class__.__qualname__}'")

    @contextmanager
    def session_cm(self):
        """Create a Session as a context manager (that can be used in ``with`` statements).

        The Session will be closed safely when the context manager exits (even in case of error).

        Example:
            You can use the context manager like this: ::

                with alchemy.session_cm() as session:
                    # Do some stuff
                    ...
                    # Commit the session
                    session.commit()

        """
        session = self.Session()
        try:
            yield session
        except Exception:
            session.rollback()
            raise
        finally:
            session.close()

    @asynccontextmanager
    async def session_acm(self):
        """Create a Session as a async context manager (that can be used in ``async with`` statements).

        The Session will be closed safely when the context manager exits (even in case of error).

        Example:
            You can use the async context manager like this: ::

                async with alchemy.session_acm() as session:
                    # Do some stuff
                    ...
                    # Commit the session
                    await asyncify(session.commit)"""
        session = await asyncify(self.Session)
        try:
            yield session
        except Exception:
            session.rollback()
            raise
        finally:
            session.close()



# Imports go here!
from .api_royalnet_version import ApiRoyalnetVersionStar
from .api_login_royalnet import ApiLoginRoyalnetStar
from .api_token_info import ApiTokenInfoStar
from .api_token_passwd import ApiTokenPasswdStar
from .api_token_create import ApiTokenCreateStar
from .api_user_get import ApiUserGetStar
from .api_user_list import ApiUserListStar
from .api_user_find import ApiUserFindStar
from .api_alias_list import ApiAliasListStar
from .docs import DocsStar

# Enter the PageStars of your Pack here!
available_page_stars = [
    ApiRoyalnetVersionStar,
    ApiLoginRoyalnetStar,
    ApiTokenInfoStar,
    ApiTokenPasswdStar,
    ApiTokenCreateStar,
    ApiUserGetStar,
    ApiUserListStar,
    ApiUserFindStar,
    ApiAliasListStar,
    DocsStar,
]

# Don't change this, it should automatically generate __all__
__all__ = [star.__name__ for star in available_page_stars]



from ..errors import SerfError


class DiscordSerfError(SerfError):
    """Base class for all :mod:`royalnet.serf.discord` errors."""


class VoicePlayerError(DiscordSerfError):
    """Base class for all :class:`VoicePlayer` errors."""


class AlreadyConnectedError(VoicePlayerError):
    """Base class for the "Already Connected" errors."""


class PlayerAlreadyConnectedError(AlreadyConnectedError):
    """The :class:`VoicePlayer` is already connected to voice.

    Access the :class:`discord.VoiceClient` through :attr:`VoicePlayer.voice_client`!"""


class GuildAlreadyConnectedError(AlreadyConnectedError):
    """The :class:`discord.Client` is already connected to voice in a channel of this guild."""


class OpusNotLoadedError(VoicePlayerError):
    """The Opus library hasn't been loaded `as required
    <https://discordpy.readthedocs.io/en/latest/api.html#discord.VoiceClient>` by :mod:`discord`."""


class DiscordTimeoutError(VoicePlayerError):
    """The websocket didn't get a response from the Discord voice servers in time."""


class PlayerNotConnectedError(VoicePlayerError):
    """The :class:`VoicePlayer` isn't connected to the Discord voice servers.

    Use :meth:`VoicePlayer.connect` first!"""


class PlayerAlreadyPlayingError(VoicePlayerError):
    """The :class:`VoicePlayer` is already playing audio and cannot start playing audio again."""



from typing import *
import asyncio as aio
import datetime
import dateparser
import logging
import royalnet.utils as ru
import youtube_dl


log = logging.getLogger(__name__)


class YtdlInfo:
    """A wrapper around `youtube_dl <https://ytdl-org.github.io/youtube-dl/index.html>`_ extracted info."""

    _default_ytdl_args = {
        "quiet": True,  # Do not print messages to stdout.
        "noplaylist": True,  # Download single video instead of a playlist if in doubt.
        "no_warnings": True,  # Do not print out anything for warnings.
        "outtmpl": "./downloads/%(epoch)s-%(title)s-%(id)s.%(ext)s",  # Use the default outtmpl.
        "ignoreerrors": True  # Ignore unavailable videos
    }

    def __init__(self, info: Dict[str, Any]):
        """Create a :class:`YtdlInfo` from the dict returned by the :func:`YoutubeDL.extract_info` function.

        Warning:
            Does not download the info, to do that use :func:`.retrieve_for_url`."""
        self.id: Optional[str] = info.get("id")
        self.uploader: Optional[str] = info.get("uploader")
        self.uploader_id: Optional[str] = info.get("uploader_id")
        self.uploader_url: Optional[str] = info.get("uploader_url")
        self.channel_id: Optional[str] = info.get("channel_id")
        self.channel_url: Optional[str] = info.get("channel_url")
        self.upload_date: Optional[datetime.datetime] = dateparser.parse(ru.ytdldateformat(info.get("upload_date")))
        self.license: Optional[str] = info.get("license")
        self.creator: Optional[...] = info.get("creator")
        self.title: Optional[str] = info.get("title")
        self.alt_title: Optional[...] = info.get("alt_title")
        self.thumbnail: Optional[str] = info.get("thumbnail")
        self.description: Optional[str] = info.get("description")
        self.categories: Optional[List[str]] = info.get("categories")
        self.tags: Optional[List[str]] = info.get("tags")
        self.subtitles: Optional[Dict[str, List[Dict[str, str]]]] = info.get("subtitles")
        self.automatic_captions: Optional[dict] = info.get("automatic_captions")
        self.duration: Optional[datetime.timedelta] = datetime.timedelta(seconds=info.get("duration", 0))
        self.age_limit: Optional[int] = info.get("age_limit")
        self.annotations: Optional[...] = info.get("annotations")
        self.chapters: Optional[...] = info.get("chapters")
        self.webpage_url: Optional[str] = info.get("webpage_url")
        self.view_count: Optional[int] = info.get("view_count")
        self.like_count: Optional[int] = info.get("like_count")
        self.dislike_count: Optional[int] = info.get("dislike_count")
        self.average_rating: Optional[...] = info.get("average_rating")
        self.formats: Optional[list] = info.get("formats")
        self.is_live: Optional[bool] = info.get("is_live")
        self.start_time: Optional[float] = info.get("start_time")
        self.end_time: Optional[float] = info.get("end_time")
        self.series: Optional[str] = info.get("series")
        self.season_number: Optional[int] = info.get("season_number")
        self.episode_number: Optional[int] = info.get("episode_number")
        self.track: Optional[...] = info.get("track")
        self.artist: Optional[...] = info.get("artist")
        self.extractor: Optional[str] = info.get("extractor")
        self.webpage_url_basename: Optional[str] = info.get("webpage_url_basename")
        self.extractor_key: Optional[str] = info.get("extractor_key")
        self.playlist: Optional[str] = info.get("playlist")
        self.playlist_index: Optional[int] = info.get("playlist_index")
        self.thumbnails: Optional[List[Dict[str, str]]] = info.get("thumbnails")
        self.display_id: Optional[str] = info.get("display_id")
        self.requested_subtitles: Optional[...] = info.get("requested_subtitles")
        self.requested_formats: Optional[tuple] = info.get("requested_formats")
        self.format: Optional[str] = info.get("format")
        self.format_id: Optional[str] = info.get("format_id")
        self.width: Optional[int] = info.get("width")
        self.height: Optional[int] = info.get("height")
        self.resolution: Optional[...] = info.get("resolution")
        self.fps: Optional[int] = info.get("fps")
        self.vcodec: Optional[str] = info.get("vcodec")
        self.vbr: Optional[int] = info.get("vbr")
        self.stretched_ratio: Optional[...] = info.get("stretched_ratio")
        self.acodec: Optional[str] = info.get("acodec")
        self.abr: Optional[int] = info.get("abr")
        self.ext: Optional[str] = info.get("ext")
        # Additional custom information
        self.album: Optional[str] = None

    @classmethod
    async def from_url(cls, url, loop: Optional[aio.AbstractEventLoop] = None, **ytdl_args) -> List["YtdlInfo"]:
        """Fetch the info for an url through :class:`YoutubeDL`.

        Returns:
            A :class:`list` containing the infos for the requested videos."""
        if loop is None:
            loop: aio.AbstractEventLoop = aio.get_event_loop()
        # So many redundant options!
        log.debug(f"Fetching info: {url}")
        with youtube_dl.YoutubeDL({**cls._default_ytdl_args, **ytdl_args}) as ytdl:
            first_info = await ru.asyncify(ytdl.extract_info, loop=loop, url=url, download=False)
        # No video was found
        if first_info is None:
            return []
        # If it is a playlist, create multiple videos!
        if "entries" in first_info:
            if len(first_info["entries"]) == 0:
                return []
            if first_info["entries"][0] is None:
                return []
            log.debug(f"Found a playlist: {url}")
            second_info_list = []
            for second_info in first_info["entries"]:
                if second_info is None:
                    continue
                second_info_list.append(YtdlInfo(second_info))
            return second_info_list
        log.debug(f"Found a single video: {url}")
        return [YtdlInfo(first_info)]

    def __repr__(self):
        if self.title:
            return f"<YtdlInfo of '{self.title}'>"
        if self.webpage_url:
            return f"<YtdlInfo for '{self.webpage_url}'>"
        return f"<YtdlInfo id={self.id} ...>"

    def __str__(self):
        if self.title:
            return self.title
        if self.webpage_url:
            return self.webpage_url
        return self.id



from typing import *
import contextlib
import logging
import asyncio as aio
import royalnet.utils as ru
from .errors import UnsupportedError
from .commandinterface import CommandInterface
from royalnet.backpack.tables.aliases import Alias

if TYPE_CHECKING:
    from .keyboardkey import KeyboardKey
    from royalnet.backpack.tables.users import User

log = logging.getLogger(__name__)


class CommandData:
    def __init__(self, interface: CommandInterface, loop: aio.AbstractEventLoop):
        self.loop: aio.AbstractEventLoop = loop
        self._interface: CommandInterface = interface
        self._session = None

    # TODO: make this asyncronous... somehow?
    @property
    def session(self):
        if self._session is None:
            if self._interface.alchemy is None:
                raise UnsupportedError("'alchemy' is not enabled on this Royalnet instance")
            log.debug("Creating Session...")
            self._session = self._interface.alchemy.Session()
        return self._session

    async def session_commit(self):
        """Asyncronously commit the :attr:`.session` of this object."""
        if self._session:
            log.warning("Session had to be created to be committed")
        # noinspection PyUnresolvedReferences
        log.debug("Committing Session...")
        await ru.asyncify(self.session.commit)

    async def session_close(self):
        """Asyncronously close the :attr:`.session` of this object."""
        if self._session is not None:
            log.debug("Closing Session...")
            await ru.asyncify(self._session.close)

    async def reply(self, text: str) -> None:
        """Send a text message to the channel where the call was made.

        Parameters:
             text: The text to be sent, possibly formatted in the weird undescribed markup that I'm using."""
        raise UnsupportedError(f"'{self.reply.__name__}' is not supported")

    async def get_author(self, error_if_none: bool = False):
        """Try to find the identifier of the user that sent the message.
        That probably means, the database row identifying the user.

        Parameters:
            error_if_none: Raise an exception if this is True and the call has no author."""
        raise UnsupportedError(f"'{self.get_author.__name__}' is not supported")

    async def delete_invoking(self, error_if_unavailable=False) -> None:
        """Delete the invoking message, if supported by the interface.

        The invoking message is the message send by the user that contains the command.

        Parameters:
            error_if_unavailable: if True, raise an exception if the message cannot been deleted."""
        if error_if_unavailable:
            raise UnsupportedError(f"'{self.delete_invoking.__name__}' is not supported")

    async def find_user(self, alias: str) -> Optional["User"]:
        """Find the User having a specific Alias.

        Parameters:
            alias: the Alias to search for."""
        return await Alias.find_user(self._interface.alchemy, self.session, alias)

    @contextlib.asynccontextmanager
    async def keyboard(self, text, keys: List["KeyboardKey"]):
        yield
        raise UnsupportedError(f"{self.keyboard.__name__} is not supported")

    @classmethod
    def register_keyboard_key(cls, identifier: str, key: "KeyboardKey"):
        raise UnsupportedError(f"{cls.register_keyboard_key.__name__} is not supported")

    @classmethod
    def unregister_keyboard_key(cls, identifier: str):
        raise UnsupportedError(f"{cls.unregister_keyboard_key.__name__} is not supported")



import logging
from typing import Optional, AsyncGenerator, Tuple, Any, Dict
try:
    import discord
except ImportError:
    discord = None


log = logging.getLogger(__name__)


class Playable:
    """An abstract class representing something that can be played back in a :class:`VoicePlayer`."""
    def __init__(self):
        """Create a :class:`Playable`.

        Warning:
            Avoid using this method, as it does not initialize the generator! Use :meth:`.create` instead."""
        log.debug("Creating a Playable...")
        self.generator: Optional[AsyncGenerator[Optional["discord.AudioSource"],
                                                Tuple[Tuple[Any, ...], Dict[str, Any]]]] = self._generator()

    # PyCharm doesn't like what I'm doing here.
    # noinspection PyTypeChecker
    @classmethod
    async def create(cls, *args, **kwargs):
        """Create a :class:`Playable` and initialize its generator."""
        playable = cls(*args, **kwargs)
        log.debug("Sending None to the generator...")
        await playable.generator.asend(None)
        log.debug("Playable ready!")
        return playable

    async def next(self, *args, **kwargs) -> Optional["discord.AudioSource"]:
        """Get the next :class:`discord.AudioSource` that should be played.

        Called when the :class:`Playable` is first attached to a :class:`VoicePlayer` and when a
        :class:`discord.AudioSource` stops playing.

        Args and kwargs can be used to pass data to the generator.

        Returns:
             :const:`None` if there is nothing available to play, otherwise the :class:`discord.AudioSource` that should
             be played.
        """
        log.debug("Getting next AudioSource...")
        audio_source: Optional["discord.AudioSource"] = await self.generator.asend((args, kwargs,))
        log.debug(f"Next: {audio_source}")
        return audio_source

    async def _generator(self) \
            -> AsyncGenerator[Optional["discord.AudioSource"], Tuple[Tuple[Any, ...], Dict[str, Any]]]:
        """Create an async generator that returns the next source to be played;
        it can take a args+kwargs tuple in input to optionally select a different source.

        Note:
            For `weird Python reasons
            <https://www.python.org/dev/peps/pep-0525/#support-for-asynchronous-iteration-protocol>`, the generator
            should ``yield`` once before doing anything else."""
        yield
        raise NotImplementedError()

    async def destroy(self):
        """Clean up the Playable, as it is about to be replaced or deleted."""
        raise NotImplementedError()



import bcrypt
from sqlalchemy import Column, \
                       Integer, \
                       String, \
                       LargeBinary
from sqlalchemy.ext.declarative import declared_attr


# noinspection PyAttributeOutsideInit
class User:
    __tablename__ = "users"

    @declared_attr
    def uid(self):
        return Column(Integer, unique=True, primary_key=True)

    @declared_attr
    def username(self):
        return Column(String, unique=True, nullable=False)

    @declared_attr
    def password(self):
        return Column(LargeBinary)

    @declared_attr
    def role(self):
        return Column(String, nullable=False)

    @declared_attr
    def avatar(self):
        return Column(LargeBinary)

    def json(self):
        return {
            "uid": self.uid,
            "username": self.username,
            "role": self.role,
            "avatar": self.avatar
        }

    def set_password(self, password: str):
        byte_password: bytes = bytes(password, encoding="UTF8")
        self.password = bcrypt.hashpw(byte_password, bcrypt.gensalt(14))

    def test_password(self, password: str):
        if self.password is None:
            raise ValueError("No password is set")
        byte_password: bytes = bytes(password, encoding="UTF8")
        return bcrypt.checkpw(byte_password, self.password)

    def __repr__(self):
        return f"<{self.__class__.__qualname__} {self.username}>"

    def __str__(self):
        return self.username



import royalnet.version
from royalnet.commands import *


class VersionCommand(Command):
    name: str = "version"

    description: str = "Display the current Royalnet version."

    async def run(self, args: CommandArgs, data: CommandData) -> None:
        # noinspection PyUnreachableCode
        if __debug__:
            message = f"ℹ️ Royalnet [url=https://github.com/Steffo99/royalnet/]Unreleased[/url]\n"
        else:
            message = f"ℹ️ Royalnet [url=https://github.com/Steffo99/royalnet/releases/tag/{royalnet.version.semantic}]{royalnet.version.semantic}[/url]\n"
        if "69" in royalnet.version.semantic:
            message += "(Nice.)"
        await data.reply(message)



"""The subpackage providing all functions and classes that handle the webserver and the webpages.

It requires the ``constellation`` extra to be installed (:mod:`starlette`).

You can install it with: ::

    pip install royalnet[constellation]

It optionally uses the ``sentry`` extra for error reporting.

You can install them with: ::

    pip install royalnet[constellation,sentry]

"""

from .constellation import Constellation
from .star import Star
from .pagestar import PageStar

__all__ = [
    "Constellation",
    "Star",
    "PageStar",
]
